//
// Hilfsprogramme fr den DCF-77 Dekoder
//


.extern DCF_Fehler	// wenn ein Fehler Auftritt, wird eine Zahl<>0 gespeichert


// ***** 64 bit Schieberegister um eine Position weiterschieben *****
//
//	In dem Schieberegister werden die DCF77 Datenbits gesammelt
//	Parameter 1 wird in r24:r25 bergeben und enthlt die Schieberegisteradresse
//	Parameter 2 wird in r22:R23 bergeben und enthlt das Bit was eingeschoben werden soll
//  Die Parameter sind Word-"aligned" d.h. LSB immer in der geraden Registernummer.
//	Bei 8-Bit und 16-Bit Parametern somit beginnend bei r24 dann r22 dann r20 etc.
//	Rueckgabewert in r24:r25
//	r2-r17, r28-r29 mssen gerettet werden!
//
.global schiebe
.func schiebe

schiebe:

	movw r30, r24		// Schieberegisteradresse nach Z laden
	tst 22				// einzuschiebendes Bit auf zero testen
	ldi r22, 8			// Wiederholungsanzahl laden
	mov r1, r22			// und in R1 ablegen
	clc					// carry lschen
	breq marke			// wenn einzuschiebendes Bit = zero, dann "0" einschieben
	sec					// sonst carry setzen und damit "1" einschieben
	
	marke:
		ld r22, Z			// Arbeitsregister mit Schieberegisterstelle laden
		ror r22				// carry einschieben und bertrag nach carry
		st Z+, r22			// Wert zurck in den Speicher und Zeiger auf die nchste Schieberegisterstelle
		dec r1				// Wiederholungszhler um 1 verringern
	brne marke				// zurck zum Schleifenanfang bis 0 erreicht ist

ret

.endfunc



// ***** Umwandlung von BCD in Binrdarstellung *****
//
//	Parameter 1 (BCD-Wert) wird als Byte in r24 bernommen
//		Verfahren: Binrwert = BCD-Wert - 6 * Zehnerstelle des BCD-Wertes
//	Rckgabewert binre Zahl als Byte in r24
//	Parameter 2 wird in r22:R23 bergeben und enthlt die hchste zulssige Zahl,
//		sollte diese Zahl berschritten werden, liegt ein Fehler vor.
//		Wenn die Einerstelle > 9 ist, liegt ebenfalls ein Fehler vor.
//		Der Fehlerzustand wird in der Variablen DCF_Fehler festgehalten
//
.global bcdtobin
.func bcdtobin

bcdtobin:

	mov		r25, r24		// BCD-Wert kopieren
	andi	r25, 15			// und Einerstelle maskieren
	subi	r25, 10			// hchster zulssiger Wert fr Einerstelle + 1
	brsh	zugross			// gehe zur Fehlerbearbeitung wenn R25 >= 10

	mov		r25, r24		// BCD-Wert kopieren
	swap	r25				// durch 16 teilen
	andi	r25, 15			// und Zehnerstelle mit 0F maskieren
	lsl		r25				// Zehnerstelle * 2
	sub		r24, r25		// zwischenwert = BCD - 2 * Zehnerstelle
	lsl		r25				// Zehnerstelle * 4
	sub		r24, r25		// binr = zwischenwert - 4 * Zehnerstelle
	
	sub		r22, r24		// Maximum - Binrwert >= 0 !
	brsh	ende			// gehe zum Ende, wenn Maximum >= Binrwert, andernfalls liegt eine Bereichsberschreitung vor
	
	zugross:
		sts DCF_Fehler, r22	// Fehlerzustand merken

	ende:
ret

.endfunc




// ***** Prfung der Paritt *****
//
//	Parameter 1 wird in r24:r25 bergeben und enthlt die Schieberegisteradresse
//	Parameter 2 wird in r22:R23 bergeben und enthlt die Anzahl des Bytes, die zu prfen sind
//	Rueckgabewert in r24 0=gerade Paritt, 1=ungerade Paritt.
//	Der Fehlerzustand wird in der Variablen DCF_Fehler festgehalten.
//	Wenn die Paritt ungerade ist, wird in DCF_Fehler eine 1 abgelegt.
//
//
.global paritaet
.func paritaet


paritaet:

	movW r30, r24			// Schieberegisteradresse nach Z laden
	ldi r24,0				// Ergebnis zurcksetzen
	
	byteloop:	
		ldi r25, 8			// Bitzhler = 8 
		ld r23, Z+			// aktuelles Byte nach R23
		
		bitloop:
			lsr	r23				// Bit nach CY schieben 
			adc	r24, r25		// "1"-Bits aus CY addieren (r25 ist im Mittel gerade)
			dec r25				// Bitzhler um 1 verringern
		brne bitloop			// zurck zum Schleifenanfang bis 0 erreicht ist
		
		dec r22				// Byteanzahl verringern
	brne byteloop			// zurck zum Schleifenanfang bis 0 erreicht ist
	
	andi r24,1				// niederwertigste Stelle (gerade/ungerade) maskieren
	
	breq fertig				// bei gerader Paritt zum Ende gehen
		sts DCF_Fehler, r24	// andernfalls Fehlerzustand merken
	fertig:
	
ret

.endfunc



.end
