//
// Hilfsprogramme fr die Funkfernsteuerung von Steckdosen
//
#include <avr/io.h>  
.extern VAr;



// ***** Eine Portleitung einlesen *****
//	Das Programm liest das Signal einer Portleitung ein.
//	Parameter 1 wird in r24 bergeben und enthlt die Portleitung (A0..D7)
//	Rueckgabewert: 0= '0' gelesen, 128 = '1' gelesen
//	r2-r17, r28-r29 mssen gerettet werden! (gilt beim AVRGCC immer!)
//
//
.global read_port
.func read_port
	
read_port:
	rcall	registeradresse		// Portadresse nach Z, Portinhalt nach r24, Bitmaske nach r23
	and		r24, r23			// Bit maskieren
	breq	m2					// wenn null, dann null zurckgeben
	ldi		r24, 128			// sonst 128 zurckgeben
m2:
	ret

.endfunc



// ***** Eine Portleitung schalten *****
//	Das Programm schaltet eine Portleitung ein oder aus.
//	Parameter 1 wird in r24 bergeben und enthlt den Port und die Leitung (A0, .. D7)
//  Parameter 2 wird in r22 bergeben und enthlt den Schaltcode (0=aus, 128=ein)
//	r2-r17, r28-r29 mssen gerettet werden! (gilt beim AVRGCC immer!)
//
//
.global schalte_port
.func schalte_port

schalte_port:
	rcall	registeradresse
	tst		r22			// 0 oder 1 schreiben ?
	breq	m3
	or		r24, r23	// Bit an der richtigen Stelle setzen
	rjmp	m4
m3:
	com		r23			// Maske invertieren
	and		r24, r23	// Bit lschen
m4:
	st		Z, r24		// Bits an den Port ausgeben
	ret

.endfunc



// ***** Unterprogramm zur Ermittlung der Portadresse aus dem BCD-Text *****
// Aus der Portleitung (BCD) in r24 wird die Registeradresse berechnet 
//  und in r30,r31 (=Z) zurckgegeben.
// Der Inhalt des Ports wird gelesen und in r24 zurckgegeben.
// Die Bitmaske fr die benutzte Leitung (0..7) wird in r23 abgelegt.
//
registeradresse:
	mov		r21, r24	// Portleitung merken (A0..A7, .., D0..D7)
	swap	r24			// Port ins lo nibble bringen
	com		r24			// Aus dem Hexwert A, B, C, D wird die Registeradresse $3B, $38, $35 und $32 berechnet
	andi	r24, 0x0f	// Port A $3B, Port B $38, Port C $35, Port D $32
	subi	r24, 2
	mov		r25, r24
	add		r24, r25
	add		r24, r25
	inc		r24
	inc		r24
	ori		r24, 0x30	// Registeradresse in r24 fertiggestellt
	
	ldi		r23, 1		// Bitmaske '= 2 hoch "Leitung"' errechnen
	andi	r21, 7		// Leitung maskieren, 0..7 zulssig
	breq	m1			// bei Leitung 0 sind wir bereits fertig
loop1:
	lsl		r23			// Bitmaske um ein Bit nach links schieben
	dec		r21			// und Zhler verringern
	brne	loop1		// bis wir bei null sind
m1:						// Die Bitmaske steht nun in r23
	eor		r25, r25
	movw	r30, r24	// Registeradresse nach Z bringen
	ld		r24, Z		// Portinhalt lesen und in r24 zurckgeben
	ret	



// ***** Telegramm umcodieren *****
//	Das Programm codiert 4 Byte von der Zeichendarstellung (0,1,F) in Doppelbits
//	Parameter 1 wird in r24:r25 bergeben und enthlt den Zeiger auf das Eingangs-Datenfeld
//   Hier stehen ASCII-Zeichen '0', '1', 'f', 'F', alles Andere wird als unzulssig interpretiert
//	Rueckgabewert: codiertes Byte
//	r2-r17, r28-r29 mssen gerettet werden! (gilt beim AVRGCC immer!)
//
//
.global Byte2Dbit
.func Byte2Dbit

Byte2Dbit:

	movw	r30, r24			// Adresse des Eingangsfeldes nach Z laden 
	
	ldi		r23, 4				// r23 zhlt die DBits pro Byte, hier Anfangswert setzen
	
	schleife:					// Schleife ber die vier Doppelbits pro Byte
			lsl		r24					// zwei Bit freimachen
			lsl		r24
			ld		r21, Z+				// Byte einlesen
			cpi		r21, '1'			// auf "1" testen =11
			brne	marke1				// Sprung, wenn es keine '1' war
			
			ori		r24, 3				// Es war eine '1', deshalb Bits auf 11 setzen
			rjmp	Fertig
			
		marke1:
			cpi		r21, '0'			// auf "0" testen =00
			breq	Fertig				// wenn '0', brauchen wir nichts zu tun
			
			cpi		r21, 'f'			// auf "f" testen =01
			brne	marke2				// Sprung, wenn es kein 'f' war
			
			ori		r24, 1				// Es war ein Float, deshalb Bits auf 01 setzen
			rjmp	Fertig
			
		marke2:
			cpi		r21, 'F'			// auf "F" testen =01
			brne	marke3				// Sprung, wenn es kein 'F' war
			
			ori		r24, 1				// Es war ein Float, deshalb Bits auf 01 setzen
			rjmp	Fertig
			
		marke3:
			ori		r24, 2				// alles andere gilt als fehlerhaft Code=10
			
		Fertig:
			dec		r23					// Schleifenzhler (Doppelbits pro Byte) dekrementieren
		
	brne	schleife			// Wiederholen, bis der Schleifenzhler=0 ist
	
ret
.endfunc



// ***** ein Doppelbit in ein ASCII-Zeichen umwandeln *****
//	Das Programm ...
//	Parameter 1 wird in r24:r25 bergeben und enthlt das Doppelbit
//	Rueckgabewert: ASCII-Zeichen
//	r2-r17, r28-r29 mssen gerettet werden! (gilt beim AVRGCC immer!)
//
//
.global getbyte
.func getbyte

getbyte:

	rol		r24			// Bits an die niederwertigen STellen schieben
	rol		r24
	rol		r24
	andi	r24,3		// Bits maskieren
	brne	M1	
	ldi		r24,'0'		// Bei 00 liegt eine '0' vor
	ret
M1:
	dec		r24
	brne	M2
	ldi		r24,'F'		// Bei 01 liegt ein Float vor
	ret
M2:
	dec		r24
	brne	M3
	ldi		r24,'?'		// 10 ist unzulssig
	ret
M3:
	ldi		r24,'1'		// 11 stellt eine '1' dar
	ret

.endfunc



// ***** Telegramm zur Sendermodulation erzeugen *****
//	Das Programm gibt an dem Portpin PB4 ein HX2262 kompatibles Datentelegramm aus
//	Parameter 1 wird in r24:r25 bergeben und enthlt den Zeiger auf das Datenfeld
//	Die Anzahl der Telegrammwiderholungen wird aus VAr gelesen
//	Es wird nichts zurckgegeben.
//	r2-r17, r28-r29 mssen gerettet werden!
//	Achtung: Portpin und Verzgerungszeiten sind fest codiert!
//
.global sendeTelegramm
.func sendeTelegramm

sendeTelegramm:
	lds		r22, VAr		// Anzahl der TelegrammWiederholungen aus VAr lesen
		
	Anzahlschleife:			// Schleife ber die Telegrammanzahl, Zwei Telegramme sind mindestens erforderlich
		movw	r30, r24		// Adresse des Datenfeldes nach Z laden
		ldi		r23, 3			// drei Daten-Byte = 12 Telegrammbits bearbeiten
		mov		r1, r23			// r1 zhlt die Datenbytes
		
		Byteschleife:				// Schleife ber die drei Datenbytes
			ld		r21, Z+				// Byte einlesen
			rcall	Bytebearbeitung		// Byte bearbeiten
			dec		r1					// Bytezhler decrementieren
		brne Byteschleife			// zurck zum Schleifenanfang bis 0 erreicht ist
		
		ldi		r23, 1			// Synchronimpuls ausgeben, nur ein Halbbit 
		eor		r21, r21		// r21 lschen, das Halbbit soll eine '0' sein
		rcall	synceinstieg	// und Halbbit ausgeben
		ldi		r21,7			// sieben weitere Halbbits Pause = 28 Zyklen
		mov		r1, r21			// r1 zhlt die Halbbits
		eor		r21, r21		// r21 lschen
		
		Syncpausenschleife:
			ldi		r23, 1			// ein Halbbit als Pause ausgeben
			rcall	pauseneinstieg
			dec		r1				// Zhler decrementieren
		brne	Syncpausenschleife	// zum Schleifenanfang bis 0 erreicht ist
		
		dec		r22				// Telegrammwiederholungszhler decrementieren
	brne Anzahlschleife		// zurck zum Schleifenanfang bis 0 erreicht ist
ret

	Bytebearbeitung:
		ldi		r23, 8			// acht Bit pro Byte bearbeiten (jedes Bit ist ein Halbbit des Telegramms)
		Bitschleife:			// Schleife ber die acht Bit des zu bearbeitenden Bytes
				
			synceinstieg:							// Seiteneinstieg, wenn nur der Synchronimpuls ausgegeben werden soll
				sbi		_SFR_IO_ADDR(PORTB), PB4	// Datenleitung B 4 auf 1 setzen
				
			pauseneinstieg:							// Seiteneinstieg, wenn nur die Wartetakte gebraucht werden
				rcall	einenTaktwarten
				lsl		r21							// Datenbit nach CY schieben 
			brcs	eins
				
				cbi		_SFR_IO_ADDR(PORTB), PB4	// bei Carry=0 Datenleitung auf 0 setzen ("0" bearbeiten)
				
			eins: 
				rcall	einenTaktwarten
				rcall	einenTaktwarten
				cbi		_SFR_IO_ADDR(PORTB), PB4	// Datenltg. auf 0 setzen ("1" bearbeiten, schadet aber nicht bei "0")
				rcall	einenTaktwarten
				
			dec		r23					// Bitzhler decrementieren
		brne Bitschleife		// zurck zum Schleifenanfang bis 0 erreicht ist
	ret

		einenTaktwarten:			// 370 s Warteschleife
			ldi	r27,	1293/256	// 14 Zyklen/MHz * 370 = 5180; 5180/4 = 1295;
			ldi	r26,	1293%256	// 1295 - 2 fr call/ret u.a.
			lp:
				sbiw	r26, 1			// 2 Zyklen
			brne lp						// 2 (bzw. 1)
		ret

.endfunc
.end
