/*#######################################################################################
AVR Small Webserver 

Copyright (C) 2006 Joachim Brke	Uhr mit DCF77-Betrieb

#######################################################################################*/

#include "compiler.h"
#include "uhr.h"
#include "../delay.h"
#include "../main.h"

#define TIMECOUNTER_10ms (F_CPU / TIMER_PRESCALE / 100)

char
	s1 = 0,			// Variable fr Sekunden (uptime)
	m1 = 0,			// Variable fr Minuten  (uptime)
	h1 = 0,			// Variable fr Stunden  (uptime)
	d1 = 0,			// Variable fr Tage     (uptime) bis zu 255 Tage!
	ss = 0,			// Variable fr Sekunden
	mm = 0,			// Variable fr Minuten
	hh = 0,			// Variable fr Stunden
	tt = 0,			// Variable fr Tag
	mt = 0,			// Variable fr Monat
	jj = 0,			// Variable fr Jahr
	Minute, 		// Zwischenspeicherung des dekodierten DCF-Signals
	Stunde,  		// Zwischenspeicherung des dekodierten DCF-Signals
	Tag,  			// Zwischenspeicherung des dekodierten DCF-Signals
	Monat,  		// Zwischenspeicherung des dekodierten DCF-Signals
	Jahr,  			// Zwischenspeicherung des dekodierten DCF-Signals
	Wochentag, 		// Zwischenspeicherung des dekodierten DCF-Signals
	sr[8],			// 64 bit langes Schieberegister, bestehend aus 8 Bytes. [0]=HOB, [7]=LOB. 
	status = 0,		// Programmstatus 0..4
	Bitzaehler,		// zhlt die im Sekundenabstand abgetasteten Bits
	Pegel,			// Logikpegel bei der Trgerabsenkung
	Zaehler,		// Wiederholungszhler fr timeout
	DCF_Fehler;		// Fehlermarke beim Dekodieren

 
unsigned char 
	VAS = 20;		// Schaltschwelle 0/1 fr den DCF77-Eingang, 20 fr Stromschleife, 120 bei Spannungseingang

int
	timecnt,	/* Variable mit dem Sollwert des Zeitzhlers. Wenn der Zhler diesen Wert
	erreicht oder berschritten hat, mu eine Sekunde weitergeschaltet werden. In keinem Fall 
	darf der Zhler neu gesetzt werden, weil das einen Zeitfehler zur Folge haben kann. */
	
	uptimecnt,	/* Variable mit dem Sollwert des Zeitzhlers fr die Uptime-Uhr. */

	abtast,		// Zeitpunkt der nchsten DCF-Programmaktion mit einer Signalabtastung
	Zeitpunkt;	// Merker fr Pulslngenmessung


// Assembler-Hilfsprogramme
extern void schiebe(char *sr, char bit);		// 64 bit Schieberegister
extern char bcdtobin(char bcd, char maximum);	// BCD zu Binr Wandlung mit Pfung auf Bereichsberschreitung
// der Einerstelle (0..9) und des Binrwertes durch Vergleich mit Maximum. Fehler werden in DCF-Fehler abgelegt
extern char paritaet(char *sr, char anzahl);	// Berechnung der Paritt ber "anzahl" Bytes, 
// Ein Parittsfehler wird in DCF-Fehler abgelegt


//#################################################################################################
// Diese Routine setzt den Vorteiler des 16-bit-Zhlers 1 auf 1024
//
void Clock_init (void)
//#################################################################################################
{
	TCCR1B |= (1<<CS10 | 0<<CS11 | 1<<CS12); 	// Setzen des Vorteilers von Zhler 1 auf 1024 
	VAS = 20;
// In den folgenden Variablen wird der Zhlerstand bei der Initialisierung gespeichert. 
// Spter wird hier der Sollwert abgelegt, den der Zhler fr die nchste Aktion erreicht haben mu.
	timecnt   = TCNT1;							// aktuellen Zhlerstand fr die Uhrzeit merken
	uptimecnt = TCNT1;							// aktuellen Zhlerstand fr "Uptime" merken
	abtast    = TCNT1;							// aktuellen Zhlerstand fr DCF-Abtastung merken
	// Initialisierung der LED-(PWM)-Ansteuerung an Pin 7 Port D
	LED_LINES		&= ~(1<<LED_PIN);	// Datenleitung auf "0" setzen
	LED_LINES_DIR	|= (1<<LED_PIN);	// Richtungsregister: LED_PIN = Output
	TCCR2	= 0x63;	// Phase correct PWM Mode fr OC2, CLK/32 d.h. 980 Hz Signalfrequenz bei 16 MHz Takt
	VA2		= 240;	// Startwert 26 = 10% Helligkeit
	OCR2	= VA2;	// PWM-Register auf den Startwert setzen
}


//#################################################################################################
// Diese Routine tastet das Signal am Analogeingang ab und liefert die Spannung als 8-bit-Wert zurck
//
char Signal_abtasten (void)
//#################################################################################################
{
	ADMUX  = (1<<ADLAR | 7); 					// ADC initialisieren, AREF benutzen, left adjust, Channel 7
	ADCSRA = (1<<ADEN | 1<<ADSC | 1<<ADIF); 	// ADC starten mit: enable, start conversion, Clear Interrupt Flag
	while (!(ADCSRA & (1<<ADIF))) {};		// Warten, bis die AD-Wandlung beendet ist
	return(ADCH);								// ADC Ergebnis als 8-Bit-Wert lesen und zurckgeben
}


//#################################################################################################
// Diese Routine liefert das Signal am Analogeingang als true/false zurck
// Da VREF auf der Webserverplatine mit +5V verbunden ist, enspricht 255 der Spannung 5V
// Eine Schaltschwelle VAS von 20 entspricht etwa 0,4V fr einen DCF-Empfnger mit Stromschnittstelle
// andernfalls ist 120 ein guter Wert
//
char DCF_Signal (void)
//#################################################################################################
{	
	if (Signal_abtasten() > VAS) return (0xff); else return (0);
}


//#################################################################################################
// Diese Routine prft den Zhler (Counter 1) und stellt fest, ob eine Sekunde vergangen ist 
// (d.h. timecnt wurde erreicht oder berschritten).
// Nach der Zhlung von (F_CPU/TIMER_PRESCALE) Impulsen ist eine Sekunde vergangen.
//
void Check_Clock (void)
//#################################################################################################
{
	if ((timecnt-TCNT1)>(F_CPU/TIMER_PRESCALE)) {	// Wenn timecnt berschritten wurde,
		timecnt += (F_CPU/TIMER_PRESCALE);			// geben wir den Zhlerstand fr die nchste Sekunde vor
		if (++ss == 60) {	ss = 0;					// und addieren zur Uhrzeit eine Sekunde
			if (++mm == 60) {	mm = 0;				// gfs. eine Minute
				if (++hh == 24) hh = 0;			// und gfs. eine Stunde
	}	}	}
	
	if ((uptimecnt-TCNT1)>(F_CPU/TIMER_PRESCALE)) {	// Wenn uptimecnt berschritten wurde,
		uptimecnt += (F_CPU/TIMER_PRESCALE);			// geben wir den Zhlerstand fr die nchste Sekunde vor
		if (++s1 == 60) {	s1 = 0;						// und addieren zur Uhrzeit eine Sekunde
			if (++m1 == 60) {	m1 = 0;					// gfs. eine Minute
				if (++h1 == 24) {						// und gfs. eine Stunde
					h1 = 0;
					++d1;
	}	}	}	}

	if ((abtast-TCNT1)>(F_CPU/TIMER_PRESCALE)) {	// Wenn "abtast" erreicht ist, erfordert die DCF-Uhr eine Aktion
		// Das Programm befindet sich in einem von fnf Zustnden, der mit 'status' angegeben wird
		
		if(status == 0) {					// Signalpegel feststellen
			Pegel = DCF_Signal();					// Signal abtasten und merken
			abtast += (TIMECOUNTER_10ms);			// nach 10 ms erneut abfragen
			status = 1;								// nchste Aktion: auf den ersten Signalwechsel warten
		}
		
		else 
		if(status == 1) {					// auf den ersten Signalwechsel warten
			if (Pegel != DCF_Signal())	{			// Wenn ein Signalwechsel vorliegt
				Zeitpunkt=TCNT1;					// Zeitpunkt des Signalwechsels merken
				status = 2;							// und auf den zweiten Signalwechsel warten
			}
			abtast += (TIMECOUNTER_10ms);			// nach 10 ms erneut abfragen
		}
		
		else 
		if(status == 2) {					// auf den zweiten Signalwechsel warten
			if (Pegel == DCF_Signal())	{			// Wenn der zweite Signalwechsel vorliegt, messen wir die Pulslnge
				Zeitpunkt -= TCNT1;					// Zeitdifferenz feststellen und in "Zeitpunkt" speichern
				if ((~Zeitpunkt+1)>((TIMECOUNTER_10ms)*25)) { // "Zeitpunkt" ist negativ!
					// Wenn die Differenz ber 250 ms (=1/4s) liegt, haben wir jetzt vermutlich einen Sekundenbeginn,
					// "Pegel" ist der Logikpegel fr den Trger im DCF-Signal
					abtast+=(TIMECOUNTER_10ms)*15;	// 150 ms warten
					status = 3;						// und dann Signal in der Bitmitte abtasten
					Bitzaehler = 0;					// Sekundenzhler mit 0 vorbesetzen;
				}
				else { 							// Jetzt ist das Datenbit gerade zuende und wir messen
					Pegel = ~Pegel;				// der Einfachheit halber noch einmal bis zur nchsten Signalnderung
					Zeitpunkt=TCNT1;				// Abtastzeitpunkt merken und
					abtast += (TIMECOUNTER_10ms);	// nach 10 ms Signal erneut abtasten
				}
			}
			else abtast += (TIMECOUNTER_10ms);		// noch kein Signalwechsel, nach 10 ms erneut abfragen
		}
		
		else 
		if(status == 3) {					// Signal in der Bitmitte abtasten
			if (DCF_Signal() == Pegel) 		// und mit dem Trgerpegel vergleichen
				schiebe(sr,1);					// langes Bit, d.h. "1" einschieben
				else 
				schiebe(sr,0); 				// kurzes Bit, "0" ins Schieberegister
			abtast += (TIMECOUNTER_10ms)*80;	// 800 ms warten (150 ms + 800 ms = 950 ms nach Sekundenbeginn)
			status = 4;							// auf den nchsten Sekundenbeginn warten
			Zaehler = 10;						// Wiederholungszhler auf 10 setzen
			OCR2 = VA2;							// rote LED wieder einschalten
		}
		
		else 
		if(status == 4) {					// Auf den nchsten Sekundenbeginn warten
			if (DCF_Signal() == Pegel) 		// Wenn eine Trgerabsenkung vorliegt, 
			{										// hat eine neue Sekunde angefangen
				OCR2 = 0;							// rote LED ausschalten;
				if (++Bitzaehler == 60)			// evtl. hat auch eine neue Minute angefangen
				if (!DCF_Fehler){ 					// Wenn kein DCF-Fehler vorliegt
					hh = Stunde;					// Zeit und Datum bernehmen
					mm = Minute;
					ss = 0;
					tt = Tag;
					mt = Monat;
					jj = Jahr;
					timecnt = TCNT1+(F_CPU/TIMER_PRESCALE);// und lokale (freilaufende) Uhr synchronisieren
					Bitzaehler = 0;
				}					
				abtast += (TIMECOUNTER_10ms)*15;	// 150 ms warten
				status = 3;							// und dann Signal in der Bitmitte abtasten
			}
			else {									// Wenn keine Trgerabsenkung vorliegt
				abtast += (TIMECOUNTER_10ms);		// 10 ms Wartezeit vormerken
				if (--Zaehler == 0){					// Wenn Wiederholungszhler = 0, fehlt der Sekundenbeginn
				// Mit der Voreinstellung des Wiederholungszhlers auf 10 ist das Abtastfenster 100 ms lang
					if (Bitzaehler > 37) {				// 38 Bits sollten es zum dekodieren mindestens sein!
						schiebe(sr,0);							// "0" fr 59. Sekunde einschieben
						DCF_Fehler = paritaet(&sr[0],3);		// P3 prfen und DCF_Fehler initialisieren
						// (das erste Byte ist in sr[0] und die Paritt ist ber 3 Byte zu berechnen)
						Tag		= bcdtobin(sr[2]&0x3f,31);	// Kalendertag steht nun rechtsbndig in sr[2]
						// (sr[2] ist mit 3f zu maskieren und der Wert darf 31 nicht bersteigen)
						schiebe(sr,0);							// weitere "0" einschieben
						sr[3] &= 0x7f;							// strendes Bit wegnehmen und
						paritaet(&sr[3],1);					// P2 prfen (1 Byte), Ergebnis landet in DCF-Fehler
						// (das erste Byte ist in sr[3] und es ist ein Byte zu betrachten)
						Stunde	= bcdtobin(sr[3]&0x3f,23);	// Stunde feststellen und Bereich prfen
						// (sr[3] ist mit 3f zu maskieren und der Wert darf 23 nicht bersteigen)
						Monat	= bcdtobin(sr[1]&0x1f,12);	// Kalendermonat feststellen und Bereich prfen
						paritaet(&sr[4],1);					// P1 prfen (1 Byte)
						Minute	= bcdtobin(sr[4]&0x7f,59);	// Minute feststellen und Bereich prfen
						for(status=0; status<5; status++) 
							schiebe(sr,0);						// 5 mal "0" einschieben
						Jahr	= bcdtobin(sr[1],99);			// KALENDERJAHR ist nun rechtsbndig
						Wochentag = sr[2]&0x7;					// und der WOCHENTAG ebenfalls
						if (Tag==0 || Wochentag==0 || Monat==0 ) DCF_Fehler=1; // Plausibilitt gegeben?
						if (DCF_Fehler) Bitzaehler = -1;		// so erreichen wir niemals die 60!
						else Bitzaehler = 59;					// Sekunde auf 59 setzen
						// wir befinden uns jetzt 50 ms nach dem Sekundenbeginn und wollen 50 ms vor
						// dem nchsten Sekundenbeginn wieder abtasten, deshalb
						abtast+=(TIMECOUNTER_10ms)*90;			// noch 900 ms bis zum Minutenbeginn warten
						status = 4;								// auf den Minutenbeginn warten
					}
					else {
						status = 0;				// es wurden zu wenige Bits empfangen, neu synchronisieren
					}
				} // Minutenbeginn
			} // keine Trgerabsenkung
		} // Statusabfrage
	} // DCF77 Auswertung

} // Ende Check_Clock
