/*****************************************************************************
*  Module Name:       uIP-AVR Port - main control loop shell
*  
*  Created By:        Louis Beaudoin (www.embedded-creations.com)
*
*  Original Release:  September 21, 2002 
*
*  Module Description:  
*  This main control loop shell provides everything required for a basic uIP
*  application using the RTL8019AS NIC
*
*   
*  November 16, 2003
*    Changed NIC interface from RTL8019 specific to general NIC calls
*    Calls the uip_arp_timer function every 10 seconds
*    
*  September 30, 2002
*    Added support for Imagecraft Compiler
*****************************************************************************/


#include <avr/pgmspace.h>
#include "uip.h"
#include "nic.h"
#include "uip_arp.h"
#include "compiler.h"
#include "services.h"
#include "usart.h"
#include "mca25.h"
#include "main.h"
#include "jb/uhr.h"

#define BUF ((struct uip_eth_hdr *)&uip_buf[0])


volatile unsigned char usart_timeout;
 
mca25_t mca25_stat;


/*****************************************************************************
*  Periodic Timout Functions and variables
*
*  The periodic timeout rate can be changed depeding on your application
*  Modify these functions and variables based on your AVR device and clock
*    rate
*  The current setup will interrupt every 256 timer ticks when the timer
*    counter overflows.  timerCounter must count until 0.5 seconds have
*    elapsed
*****************************************************************************/


static unsigned char timerCounter;
#define TIMERCOUNTER_PERIODIC_TIMEOUT (F_CPU / TIMER_PRESCALE / 256 / 2)	// = 28,125

int timeCounter;
#define TIMECOUNTER_500ms_TIMEOUT (F_CPU / TIMER_PRESCALE / 2)


void initTimer(void)
{
  TCCR0=5; //outp( 5, TCCR0 ) ;  // timer0 prescale 1/1024

  // interrupt on overflow
  TIMSK|=(1<<TOIE0); //sbi( TIMSK, TOIE0 ) ;
  timerCounter = 0;
}


#ifdef __IMAGECRAFT__
#pragma interrupt_handler SIG_OVERFLOW0:iv_TIMER0_OVF
#endif

SIGNAL(SIG_OVERFLOW0)	// 56,25 Interrupts/s
{
  timerCounter++;
  if((unsigned char)!(timerCounter&3))	// jedes 4. Mal aufrufen, also 14,0625/s
	if (usart_timeout) usart_timeout--;	// Bei 180 dauert das timeout dann 12,8s
}



/*****************************************************************************
*  Main Control Loop
*
*  
*****************************************************************************/
int main(void)
{
	unsigned char i;
	unsigned char arptimer=0;

	checkEEPROMip();
	
	nic_init();			// init NIC device driver

	uip_init();			// init uIP
	
	services_init();	// init app
	
	uip_arp_init();		// init ARP cache

	initTimer();		// init periodic timer
	
	Clock_init();		// Uhr initialisieren und starten
	
	timeCounter = TCNT1 + (F_CPU/1024) + 10;	// uiP-Zeitgeber initialisieren
	
	OWI_init();			// one-wire Schnittstelle initialisieren

	SET_USART_9600();
  
	sei();

	mca25_stat.init=mca25_init();
	
	while(1) {	
		
		Check_Clock ();				// Uhr abfragen
		
		OWI_check();				// Dallas-Bausteine bearbeiten
		
		uip_len = nic_poll();		// look for a packet
		if(uip_len == 0) {
			if((timeCounter-TCNT1) > (F_CPU/1024)) {	// if timed out, call periodic function for each connection
				timeCounter += (TIMECOUNTER_500ms_TIMEOUT);	// 0,5 s
				
				for(i = 0; i < UIP_CONNS; i++) {	// call periodic function for each connection
					uip_periodic(i);
					if(uip_len > 0) {				// transmit a packet, if one is ready
						uip_arp_out();
						nic_send();
					}
				}
				
				if(++arptimer == 20) {			// Call the ARP timer function every 10 seconds.
					uip_arp_timer();
					arptimer = 0;
				}
			}
		}
		else { // packet received
			if(BUF->type == htons(UIP_ETHTYPE_IP)){			// process an IP packet
				uip_arp_ipin();		// add the source to the ARP cache
				uip_input();		// also correctly set the ethernet packet length before processing
				
				if(uip_len > 0) {	// transmit a packet, if one is ready
					uip_arp_out();
					nic_send();
				}
			}
			else if(BUF->type == htons(UIP_ETHTYPE_ARP)) {	// process an ARP packet
				uip_arp_arpin();
				
				if(uip_len > 0)	// transmit a packet, if one is ready
				nic_send();
			}
		}
	}
	return 1;
}
