/*****************************************************************************
*  Simple Application using uIP 0.9
*****************************************************************************/

#include <avr/pgmspace.h>
#include <avr/io.h>
#include <string.h>
#include "compiler.h"
#include "services.h"
#include "delay.h"

#include "main.h"
#include "mca25.h"
#include "uipopt.h"
#include "uip.h"
#include "uip_arp.h"
#include "OWI/OWIPolled.h"
#include "OWI/OWIcrc.h"
#include "jb/uhr.h"
#include "jb/tokens.h"
#include "jb/html.h"

#define HTTP_NOGET   0
#define HTTP_GET     1
#define HTTP_FILE    2
#define RAW_FILE     3
#define HTTP_GET_404 4
#define HTTP_GET_PIC 5
#define HTTP_PIC     6

// MIME-Header
#define KeinHeader		0
#define H_GIF			1
#define H_JPEG			2
#define H_HTML			3
#define H_XICON			4
#define H_fertig		5


extern mca25_t mca25_stat;
extern unsigned char mca25_buf[CAM_BUFFER_LEN];
extern void get_request(char* Zeiger);


#include "jb/interpreter-up.c"

prog_char Header_Start[] = {"HTTP/1.0 200 OK\r\nServer: AVR\r\nConnection: close\r\nContent-Type: "};

void sendraw_P(struct http_state* s) {
	// gfs. wird den Daten ein MIME-Type vorangestellt
	if (s->header) {	// MIME-Type (Multipurpose Internet Mail Extensions) den Daten voranstellen
		int headerlaenge = sizeof(Header_Start)-1;				// Stringende wird nicht bercksichtigt!
		memcpy_P((u8_t*)uip_appdata,Header_Start,headerlaenge);	// Anfang des Headers kopieren
		if (s->header==H_HTML) {		// Content-Type: text, MIME-Type: html
			memcpy_P((u8_t*)uip_appdata+headerlaenge,PSTR("text/html"),9);
			headerlaenge += 9;
		} else {						// Content-Type: image
			memcpy_P((u8_t*)uip_appdata+headerlaenge,PSTR("image/"),6);
			headerlaenge += 6;
			if (s->header==H_GIF) {	// MIME-Type: gif
				memcpy_P((u8_t*)uip_appdata+headerlaenge,PSTR("gif"),3);
				headerlaenge += 3;
			} else
			if (s->header==H_JPEG) {	// MIME-Type: jpeg
				memcpy_P((u8_t*)uip_appdata+headerlaenge,PSTR("jpeg"),4);
				headerlaenge += 4;
			} else
			if (s->header==H_XICON) {	// MIME-Type: x-icon
				memcpy_P((u8_t*)uip_appdata+headerlaenge,PSTR("x-icon"),6);
				headerlaenge += 6;
			}
		}
		// Im Falle des Headers senden wir ein Paket, dass nur den Header enthlt
		memcpy_P((u8_t*)uip_appdata+headerlaenge,PSTR("\r\n\r\n"),4);	// Abschlu des Headers kopieren
		headerlaenge += 4;
		uip_send(uip_appdata,headerlaenge);
	} // Ende header
	
	else {	// Wenn kein Header zu senden ist, werden Datenpakete gesendet
		PGM_P Lesezeiger;
		Lesezeiger= s->dataptr_P;				// Die Quelldaten beginnen bei s->dataptr_P
		Schreibzeiger = (char*)uip_appdata;	// Der Sendepuffer beginnt bei uip_appdata
		int zusendendeBytes = 0;				// Anzahl der zu sendenden Bytes
		
		if (s->state==RAW_FILE)	{		// im Falle des "raw-files" bertragen wir die maximale Paketgre
			zusendendeBytes = (s->dataleft>uip_mss())?uip_mss():s->dataleft;	// Paketgre ermitteln
			memcpy_P(Schreibzeiger,Lesezeiger,zusendendeBytes);	// Paket in den Buffer kopieren
		} 
		
		else
		if (s->state==HTTP_FILE) {		// HTTP-Files werden durch den Interpreter geschickt
			char Datum;
			do {
				Datum = pgm_read_byte(Lesezeiger++);	// Byte lesen
				if (Datum == '\033')  {	// auf <ESCAPE>-Zeichen testen
					#include "jb/interpreter.c"					
				}
				else	
				if (Datum == '\032')	 	// <SUB>-Zeichen, Token ausgeben
					token_out(get_index(Lesezeiger++));
				else	
				if (Datum == '\003')		// <ETX>-Zeichen, Dateiende
					s->header = H_fertig;
				else *Schreibzeiger++ = Datum;	// Byte unverndert schreiben
					
				zusendendeBytes = (int)Schreibzeiger - (int)uip_appdata;
			} while ((zusendendeBytes < (uip_mss()-80)) && (s->header != H_fertig));
			s->nxtdatptr_P = Lesezeiger; 
		}	// Ende der Bearbeitung von HTTP-Dateien
		
		uip_send(uip_appdata,zusendendeBytes);	// Bufferinhalt als Paket verschicken
	}
}


void start_pic_trans(struct http_state* s,unsigned int *len) {
	mca25_stat.init=mca25_configure();
	if(mca25_stat.init) {	
		// do something here - create a message for user
		uip_abort();
		return;
	} 
	if (mca25_stat.busy) {
		uip_close();
		return;
	}
	s->state=HTTP_PIC;
	mca25_stat.busy=1;
	mca25_copy_data(mca25_buf,len);
	uip_send(mca25_buf,*len);
}



void httpd(void) {
	/* This is the web server: */
	
	static unsigned int len=0;
	static char ret=1;

	struct http_state *s;
	s = (struct http_state *)uip_conn->appstate;
	
    /* We use the uip_ test functions to deduce why we were called. If uip_connected() is non-zero, we were called
       because a remote host has connected to us. If uip_newdata() is non-zero, we were called because the remote
	   host has sent us new data, and if uip_acked() is non-zero, the remote host has acknowledged the data we
       previously sent to it. */

	if(uip_connected()) {
		/* Since we have just been connected with the remote host, we reset the state for this connection. 
		 The ->state is set to HTTP_NOGET to signal that we haven't received any HTTP GET request for this
         connection yet. */
		s->state=HTTP_NOGET;
		return;
	}


	if(uip_newdata()) {
		/* This is the first data we receive, and it should contain a GET. */
		switch (s->state) {
			
			case HTTP_NOGET:
				/* If it isn't a GET, we abort the connection. */
				if (strncmp_P((char*)uip_appdata,PSTR("GET /"),5))	uip_abort();
				
				/* The web server is now no longer in the HTTP_NOGET state, but in the HTTP_FILE state
				since it has now got the GET from the client and will start transmitting the file. */
				
				get_request ((char*)uip_appdata+5);	// Daten scannen und in die Variablen bertragen
				
				s->state=RAW_FILE;
				
				if (!strncmp_P((const char*)uip_appdata+5,PSTR("cam.jpg"),7)) {
					s->state=HTTP_GET_PIC;
					start_pic_trans(s,&len);
					break;
				}
				
				if (!strncmp_P((const char*)uip_appdata+5,PSTR("t.jpg"),5)) { 
					s->dataptr_P	= Picture1;
					s->dataleft		= sizeof(Picture1);
					s->header		= H_JPEG;
					sendraw_P(s);
					break;     
				}
				
				if (!strncmp_P((const char*)uip_appdata+5,PSTR("favicon.ico"),11)) {
					s->dataptr_P	= Picture2;
					s->dataleft		= sizeof(Picture2);
					s->header		= H_XICON;
					sendraw_P(s);
					break;     
				}
				
				if (!strncmp_P((const char*)uip_appdata+5,PSTR("b.gif"),5)) {
					s->dataptr_P	= Picture3;
					s->dataleft		= sizeof(Picture3);
					s->header		= H_GIF;
					sendraw_P(s);
					break;     
				}
				
				s->header = H_HTML;
				s->state=HTTP_FILE;
				
				if ((uip_appdata[5]==' ') || (!strncmp_P((char*)uip_appdata+5,PSTR("index.htm"),9))) {
					s->dataptr_P	= Page8;
					sendraw_P(s);
					break;     
				}
				
				
				if (!strncmp_P((const char*)uip_appdata+5,PSTR("s.htm"),5)) {
					s->dataptr_P	= Page1;
					sendraw_P(s);
					break;     
				}
				
				if (!strncmp_P((const char*)uip_appdata+5,PSTR("t.htm"),5)) {
					s->dataptr_P	= Page2;
					sendraw_P(s);
					break;     
				}
				
				if (!strncmp_P((const char*)uip_appdata+5,PSTR("d.htm"),5)) {
					s->dataptr_P	= Page3;
					sendraw_P(s);
					break;     
				}
				
				if (!strncmp_P((const char*)uip_appdata+5,PSTR("e.htm"),5)) {
					s->dataptr_P	= Page4;
					sendraw_P(s);
					break;     
				}
				
				if (!strncmp_P((const char*)uip_appdata+5,PSTR("n.htm"),5)) {
					s->dataptr_P	= Page5;
					sendraw_P(s);
					break;     
				}
				
				if (!strncmp_P((const char*)uip_appdata+5,PSTR("log.htm"),7)) {
					s->dataptr_P	= Page6;
					sendraw_P(s);
					break;     
				}
				
				if (!strncmp_P((const char*)uip_appdata+5,PSTR("c.htm"),5)) {
					s->dataptr_P	= Page7;
					sendraw_P(s);
					break;     
				}
				
				if (!strncmp_P((const char*)uip_appdata+5,PSTR("m.htm"),5)) {
					s->dataptr_P	= Page9;
					sendraw_P(s);
					break;     
				}
				
				{
					s->dataptr_P	= Page0;
					sendraw_P(s);
					break;     
				}
		}	// Ende switch
	}	// Ende uip_newdata


	if(uip_rexmit()) {					// retransmit
		switch (s->state) {
			case RAW_FILE:		sendraw_P(s);				break;
			case HTTP_FILE:	sendraw_P(s);				break;
			case HTTP_PIC:		uip_send(mca25_buf,len);	break;
		}
		return;
	}


	if(uip_acked()) {					// besttigt, nchster Block
    /* Check if the client (remote end) has acknowledged any data that we've previously sent. If so, we move the file
	 pointer further into the file and send back more data. If we are out of data to send, we close the connection. */
		switch (s->state) {
		
		case RAW_FILE:
			if (s->header) s->header = KeinHeader; // Falls ein Header gesendet wurde, brauchen wir keinen Weiteren
			else {
				if(s->dataleft < uip_mss()) {		// Falls das letzte Paket nicht die maximale Gre hatte
					uip_close();					// ... ist die bertragung nun beendet
					break;
				}
				s->dataptr_P	+= uip_conn->len;	// Datenanfang weitersetzen
				s->dataleft		-= uip_conn->len;	// Anzahl verbliebender Bytes entsprechend reduzieren
			}
			sendraw_P(s);						// nchstes Paket senden
			break;
			
		case HTTP_FILE:
			if (s->header && (s->header != H_fertig)) s->header = KeinHeader; // Falls ein Header gesendet wurde, brauchen wir keinen Weiteren
			else {
				if (s->header == H_fertig) {		// Falls es das letzte Paket war
					uip_close();					// ... ist die bertragung nun beendet
					break;
				}
				s->dataptr_P = s->nxtdatptr_P; 	// Datenanfang weitersetzen
			}
			sendraw_P(s);						// nchstes Paket senden
			break;
			
		case HTTP_PIC:
			if (ret) { 
				ret=mca25_copy_data(mca25_buf,&len);
				uip_send(mca25_buf,len);
				return;  
			}
			ret=1;
			mca25_stat.busy=0;
			uip_close(); 
			break; 
		}
	}


	if((uip_aborted()||uip_timedout()) && mca25_stat.busy) {
		while (mca25_copy_data(mca25_buf,&len)); 
		mca25_stat.busy=0;
		ret=1;
	}
    /* Finally, return to uIP. Our outgoing packet will soon be on its way... */

}



void services_init(void) {
   
	uip_listen(HTONS(80));
}



void services_main(void) {

	switch(uip_conn->lport) {

	case HTONS(80):
		httpd();
		break;

	default:
		uip_abort();
		break;
	
	}  
}
